import xarray as xr
import nexusformat.nexus as nx

def to_datarr(nxdata):
    ''' Convert NeXus NXdata to an xarray DataArray
    
    This function convert the NXdata group to an
    xarray DataArray, keeping the @default signal, all the axes
    and attributes.
    
    Arguments
        nxdata: NeXus NXdata to convert
    
    Returns:
        xarray DataArray
    
    Example:
        import nxarray as nxr
        
        nxdata = nexus.NXdata()
        dr = nxr.to_datarr(nxdata)
    '''

    ## Retrieve data and name
    data = nxdata.nxsignal.nxdata
    name = nxdata.nxsignal.nxname

    ## Generate coordinates dictionary
    coords = {}
    for axes in nxdata.nxaxes:
        coords[axes.nxname] = axes.nxdata

    ## Retrieve dimensions
    dims = nxdata.attrs["axes"]

    ## Add NXdata and NXsignal attributes to the DataArray
    attrs = {**_get_attrs(nxdata),
             **_get_attrs(nxdata.nxsignal)}

    # Create the xarray DataArray
    datarr = xr.DataArray(data,
                          name=name,
                          coords=coords,
                          dims=dims,
                          attrs=attrs)

    return datarr

def to_datset(nxentry):
    ''' Convert NeXus NXentry to an xarray Dataset
    
    This function convert the NXdata groups in the NXentry
    to DataArrays of an xarray Dataset.
    Other groups in the NXentry are saved in the Dataset attribute "NX" as a dictionary.
    
    Arguments
        nxentry: NeXus NXentry to convert
    
    Returns:
        xarray Dataset
    
    Example:
        import nxarray as nxr
        
        nxentry = nexus.NXentry()
        ds = nxr.to_datset(nxentry)
    '''

    ## Initialize Dataset
    ds = xr.Dataset()

    ## Add NeXus objects to the Dataset
    ds.attrs["NX"] = dict()
    for nxname, nxobject in nxentry.entries.items():
        if isinstance(nxobject, nx.NXdata):
            # Add NXdata groups as DataArrays
            ds[nxname] = to_datarr(nxobject)
        else:
            # Retrieve other NeXus groups in a dictionary
            ds.attrs["NX"][nxname] = nxobject

    ## Add NXentry attributes to the Dataset
    attrs = _get_attrs(nxentry)

    return ds

def _get_attrs(nxfield):
    ''' Convert dictionary of NXattr to a common dictionary
    '''

    # Initialize attributes dictionary
    attrs = {}

    # Loop over NXattr dictionary
    # skipping some attributes specific to NXfield
    for k,v in nxfield.attrs.items():
        try:
            axes_indices = list("{}_indices".format(a) for a in nxfield.nxaxes)
        except TypeError:
            axes_indices = list()
        if k not in ["signal", "axes", "default"] + axes_indices:
            attrs[k] = v.nxvalue

    return attrs

def load(filename, entry=None):
    ''' Load a NeXus file to an xarray Dataset
    
    This function load the NXdata groups in the given NXentry
    of the NeXus file and return them as an xarray Dataset.
    Other groups in the NXentry are saved in the Dataset attribute "NX" as a dictionary.
    Only one NXentry can be loaded from the NeXus file (by default the @default one).
    
    Arguments
        filename: file path to the NeXus file
        entry (optional): name of the NXentry to be loaded. If None the @default NXentry will be loaded.
    
    Returns:
        xarray Dataset
    
    Example:
        import nxarray as nxr
        
        ds = nxr.load(path/to/file.nx)
    '''

    # Open NeXus file
    f = nx.nxload(filename)

    # Get the nxentry and return it as an xarray dataset
    if entry:
        nxentry = f[entry]
    else:
        # Get the nxentry relative to @default nxdata
        nxdata = f.plottable_data
        nxentry = nxdata.nxgroup

    return to_datset(nxentry)



import nexusformat.nexus as nx
import xarray as xr

@xr.register_dataarray_accessor("nxr")
class _nxrDataArray:
    '''nxarray class extending xarray DataArray
    '''

    def __init__(self, xarray_dataarray):
        self._datarr = xarray_dataarray

    def to_nxdata(self):
        ''' Convert xarray DataArray to NeXus NXdata
        
        Returns:
            NeXus NXdata
        
        Example:
            import nxarray as nxr
            
            dr = xarray.DataArray()
            nxdata = dr.nxr.to_nxdata()
        '''

        ## Initialize NXdata
        nxname = self._datarr.name if self._datarr.name != None else "data"
        nxdata = nx.NXdata(name=nxname)

        ## Add dataset to NXdata
        self._add_to(nxdata)

        return nxdata

    def _add_to(self, nxdata):
        ''' Add DataArray to NXdata as signal with axes
        '''

        ## Add signal
        self._add_signal(nxdata)

        ## Add axes
        self._add_axes(nxdata)

        ## Add DataArray attributes to signal
        _add_attrs(self._datarr.attrs, nxdata.nxsignal)

    def _add_signal(self, nxdata):
        signal_name = self._datarr.name if self._datarr.name != None else "signal"
        nxdata.nxsignal = nx.NXfield(self._datarr.values, name=signal_name)
    
    def _add_axes(self, nxdata):
        for coord_name in self._datarr.coords:
            axes = self._datarr.coords[coord_name].values
            nxdata.nxaxes = nx.NXfield(axes, name=coord_name)
        # Overwrite the list of all xarray coordinates with just xarray dimensions
        nxdata.attrs["axes"] = list(self._datarr.dims)

    def save(self, filename, **kwargs):
        ''' Save xarray DataArray to NeXus file
        
        Arguments:
            filename: file path to .nx file
            **kwargs: any optional argument accepted by NXdata.save() method
        
        Returns:
            nothing
        
        Example:
            import nxarray as nxr
            
            dr = xarray.DataArray()
            dr.nxr.save(path/to/file.nx)
        '''

        ## Initialize NXdata
        nxdata = self._datarr.nxr.to_nxdata()

        ## Save to file
        nxdata.save(filename, **kwargs)

@xr.register_dataset_accessor("nxr")
class _nxrDataset:
    '''nxarray class extending xarray Dataset
    '''

    def __init__(self, xarray_dataset):
        self._datset = xarray_dataset

    def to_nxentry(self):
        ''' Convert xarray Dataset to NeXus NXentry
        
        Returns:
            NeXus NXentry
        
        Example:
            import nxarray as nxr
            
            ds = xarray.Dataset()
            nxentry = ds.nxr.to_nxentry()
        '''

        ## Initialize NXentry
        nxentry = nx.NXentry()

        for name, datarr in self._datset.data_vars.items():
            nxentry[name] = datarr.nxr.to_nxdata()

        ## Add dataset attributes to NXentry
        _add_attrs(self._datset.attrs, nxentry)

        # Add any other NeXus groups to the dataset
        try:
            for nxname, nxobject in self._datset.attrs["NX"].items():
                nxentry[nxname] = nxobject
        except KeyError:
            pass

        return nxentry

    def save(self, filename, **kwargs):
        ''' Save xarray Dataset to NeXus file
        
        Arguments:
            filename: file path to .nx file
            **kwargs: any optional argument accepted by NXentry.save() method
        
        Returns:
            nothing
        
        Example:
            import nxarray as nxr
            
            ds = xarray.Dataset()
            ds.nxr.save(path/to/file.nx)
        '''

        ## Initialize NXentry
        nxentry = self._datset.nxr.to_nxentry()

        ## Save to file
        nxentry.save(filename, **kwargs)

def _add_attrs(attrs, nxfield):
    for k,v in attrs.items():
        # Avoid any overwriting of specific NXfield attributes
        try:
            axes_indices = list("{}_indices".format(a) for a in nxfield.nxaxes)
        except TypeError:
            axes_indices = list()
        if k not in ["signal", "axes", "default", "NX"] + axes_indices:
            nxfield.attrs[k] = v
